package cz.muni.fi.pv112.cv4;

import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.imageio.ImageIO;
import org.joml.Matrix3f;
import org.joml.Matrix4f;
import org.joml.Vector3f;
import org.lwjgl.BufferUtils;
import org.lwjgl.Version;
import org.lwjgl.glfw.GLFWErrorCallback;
import org.lwjgl.glfw.GLFWVidMode;
import org.lwjgl.opengl.GL;
import org.lwjgl.system.MemoryStack;
import static org.lwjgl.glfw.Callbacks.glfwFreeCallbacks;
import static org.lwjgl.glfw.GLFW.*;
import static org.lwjgl.opengl.GL11.*;
import static org.lwjgl.opengl.GL12.*;
import static org.lwjgl.opengl.GL13.*;
import static org.lwjgl.opengl.GL14.*;
import static org.lwjgl.opengl.GL15.*;
import static org.lwjgl.opengl.GL20.*;
import static org.lwjgl.opengl.GL30.*;
import static org.lwjgl.system.MemoryStack.stackPush;
import static org.lwjgl.system.MemoryUtil.NULL;

public class Cv4 {

    private static final int SIZEOF_AXES_VERTEX = 6 * Float.BYTES;
    private static final int COLOR_OFFSET = 3 * Float.BYTES;

    private static final float AXES[] = {
        // .. position .......... color ....
        // x axis
        1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f,
        0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f,
        // y axis
        0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f,
        0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f,
        // z axis
        0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f,
        0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f
    };

    private static final int SIZEOF_MODEL_VERTEX = 8 * Float.BYTES;
    private static final int NORMAL_OFFSET = 3 * Float.BYTES;
    private static final int TEXCOORD_OFFSET = 6 * Float.BYTES;

    private Camera camera;

    // the window handle
    private long window;

    // window size
    private int width;
    private int height;
    private boolean resized = false;

    // animation
    private boolean animate = false;
    private float t = 0f;

    // rendering mode
    private int mode = GL_FILL;

    // model
    private ObjLoader cube;

    // our OpenGL resources
    private int axesBuffer;
    private int cubeBuffer;
    private int teapotBuffer;
    private int axesArray;
    private int cubeArray;
    private int teapotArray;

    private int woodTexture;
    private int rocksTexture;
    private int[] diceTextures = new int[6];

    // our GLSL resources
    private int axesProgram;
    private int axesAspectUniformLoc;
    private int axesLengthUniformLoc;
    private int axesMvpUniformLoc;

    private int modelProgram;
    private int modelMvpLoc;
    private int modelNLoc;
    private int modelModelLoc;

    private int lightPositionLoc;
    private int lightAmbientColorLoc;
    private int lightDiffuseColorLoc;
    private int lightSpecularColorLoc;

    private int materialAmbientColorLoc;
    private int materialDiffuseColorLoc;
    private int materialSpecularColorLoc;
    private int materialShininessLoc;

    private int eyePositionLoc;

    private int woodTexLoc;
    private int rocksTexLoc;
    private int diceTexLoc;

    public static void main(String[] args) {
        new Cv4().run();
    }

    public void run() {
        System.out.println("Hello LWJGL " + Version.getVersion() + "!");

        camera = new Camera();

        initGLFW();
        loop();

        // Free the window callbacks and destroy the window
        glfwFreeCallbacks(window);
        glfwDestroyWindow(window);

        // Terminate GLFW and free the error callback
        glfwTerminate();
        glfwSetErrorCallback(null).free();
    }

    private void initGLFW() {
        // Setup an error callback. The default implementation
        // will print the error message in System.err.
        GLFWErrorCallback.createPrint(System.err).set();

        // Initialize GLFW. Most GLFW functions will not work before doing this.
        if (!glfwInit()) {
            throw new IllegalStateException("Unable to initialize GLFW");
        }

        // Configure GLFW
        glfwDefaultWindowHints(); // optional, the current window hints are already the default
        glfwWindowHint(GLFW_VISIBLE, GLFW_FALSE); // the window will stay hidden after creation
        glfwWindowHint(GLFW_RESIZABLE, GLFW_TRUE); // the window will be resizable
        glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3); // request the OpenGL 3.3 core profile context
        glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
        glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

        // set initial width and height
        width = 640;
        height = 480;

        // Create the window
        window = glfwCreateWindow(width, height, "Hello World!", NULL, NULL);
        if (window == NULL) {
            throw new RuntimeException("Failed to create the GLFW window");
        }

        // Setup a key callback. It will be called every time a key is pressed, repeated or released.
        glfwSetKeyCallback(window, (window, key, scancode, action, mods) -> {
            if (action == GLFW_RELEASE) {
                switch (key) {
                    case GLFW_KEY_ESCAPE:
                        glfwSetWindowShouldClose(window, true); // We will detect this in the rendering loop
                        break;
                    case GLFW_KEY_A:
                        animate = !animate;
                        break;
                    case GLFW_KEY_T:
                        // TODO toggle fullscreen
                        break;
                    case GLFW_KEY_L:
                        mode = GL_LINE;
                        break;
                    case GLFW_KEY_F:
                        mode = GL_FILL;
                        break;
                }
            }
        });

        glfwSetMouseButtonCallback(window, (window, button, action, mods) -> {
            if (action == GLFW_PRESS) {
                if (button == GLFW_MOUSE_BUTTON_1) {
                    camera.updateMouseButton(Camera.Button.LEFT, true);
                } else if (button == GLFW_MOUSE_BUTTON_2) {
                    camera.updateMouseButton(Camera.Button.RIGHT, true);
                }
            } else if (action == GLFW_RELEASE) {
                if (button == GLFW_MOUSE_BUTTON_1) {
                    camera.updateMouseButton(Camera.Button.LEFT, false);
                } else if (button == GLFW_MOUSE_BUTTON_2) {
                    camera.updateMouseButton(Camera.Button.RIGHT, false);
                }
            }
        });

        glfwSetCursorPosCallback(window, (window, xpos, ypos) -> {
            camera.updateMousePosition(xpos, ypos);
        });

        // add window size callback
        glfwSetWindowSizeCallback(window, (window, width, height) -> {
            this.width = width;
            this.height = height;
            resized = true;
        });

        // Get the thread stack and push a new frame
        try (MemoryStack stack = stackPush()) {
            IntBuffer pWidth = stack.mallocInt(1); // int*
            IntBuffer pHeight = stack.mallocInt(1); // int*

            // Get the window size passed to glfwCreateWindow
            glfwGetWindowSize(window, pWidth, pHeight);

            // Get the resolution of the primary monitor
            GLFWVidMode vidmode = glfwGetVideoMode(glfwGetPrimaryMonitor());

            // Center the window
            glfwSetWindowPos(
                    window,
                    (vidmode.width() - pWidth.get(0)) / 2,
                    (vidmode.height() - pHeight.get(0)) / 2
            );
        } // the stack frame is popped automatically

        // Make the OpenGL context current
        glfwMakeContextCurrent(window);
        // Enable v-sync
        glfwSwapInterval(1);

        // Make the window visible
        glfwShowWindow(window);

        // This line is critical for LWJGL's interoperation with GLFW's
        // OpenGL context, or any context that is managed externally.
        // LWJGL detects the context that is current in the current thread,
        // creates the GLCapabilities instance and makes the OpenGL
        // bindings available for use.
        GL.createCapabilities();
    }

    private void loop() {
        // Prepare data for rendering
        init();

        // Run the rendering loop until the user has attempted to close
        // the window or has pressed the ESCAPE key.
        while (!glfwWindowShouldClose(window)) {
            render();

            glfwSwapBuffers(window); // swap the color buffers

            // Poll for window events. The key callback above will only be
            // invoked during this call.
            glfwPollEvents();
        }
    }

    private void init() {
        // empty scene color
        // Task 10: set clear color to a brighter color (e.g., 0.15, 0.15, 0.15) so that we can better see the cube
        glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        glLineWidth(3.0f); // makes lines thicker

        // Task 9:  disable depth test in order to draw also occluded geometry
        glEnable(GL_DEPTH_TEST);
        

        // Task 9:  enable blending using glEnable(GL_BLEND)
        //          set blending function by glBendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        //              it will work like this: resultFrag = a * newFrag + (1-a) * oldFrag
        
        
        // Task 10: enable face culling in order to draw back faces first and front faces afterwards
        //             use glEnable(GL_CULL_FACE)
        
        
        // load GLSL program (vertex, fragment shaders) and textures
        try {
            axesProgram = loadProgram("/resources/shaders/axes.vs.glsl",
                    "/resources/shaders/axes.fs.glsl");
            modelProgram = loadProgram("/resources/shaders/model.vs.glsl",
                    "/resources/shaders/model.fs.glsl");

            // Task 1:  load wood.jpg texture using loadTexture("/resources/textures/wood.jpg"), store it to woodTexture variable
            // Task 7:  load all dice1-6.png textures from "/resources/textures/dice1-6.png", store them to diceTextures array
            // Task 8:  load rocks.jpg texture from "/resources/textures/rocks.jpg", store it to rocksTexture variable
            
            
        } catch (IOException ex) {
            Logger.getLogger(Cv4.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(1);
        }

        // get uniform locations
        // axes program uniforms
        axesAspectUniformLoc = glGetUniformLocation(axesProgram, "aspect");
        axesLengthUniformLoc = glGetUniformLocation(axesProgram, "len");
        axesMvpUniformLoc = glGetUniformLocation(axesProgram, "MVP");

        // model program uniforms
        modelMvpLoc = glGetUniformLocation(modelProgram, "MVP");
        modelNLoc = glGetUniformLocation(modelProgram, "N");
        modelModelLoc = glGetUniformLocation(modelProgram, "model");

        lightPositionLoc = glGetUniformLocation(modelProgram, "lightPosition");
        lightAmbientColorLoc = glGetUniformLocation(modelProgram, "lightAmbientColor");
        lightDiffuseColorLoc = glGetUniformLocation(modelProgram, "lightDiffuseColor");
        lightSpecularColorLoc = glGetUniformLocation(modelProgram, "lightSpecularColor");

        eyePositionLoc = glGetUniformLocation(modelProgram, "eyePosition");

        materialAmbientColorLoc = glGetUniformLocation(modelProgram, "materialAmbientColor");
        materialDiffuseColorLoc = glGetUniformLocation(modelProgram, "materialDiffuseColor");
        materialSpecularColorLoc = glGetUniformLocation(modelProgram, "materialSpecularColor");
        materialShininessLoc = glGetUniformLocation(modelProgram, "materialShininess");

        // Task 1:  get location of woodTex uniform using glGetUniformLocation(...), store it to woodTexLoc variable
        // Task 7:  get location of diceTex uniform, store it to diceTexLoc
        // Task 8:  get location of rocksTex uniform, store it to rocksTexLoc
        

        // create buffers with geometry
        int[] buffers = new int[3];
        glGenBuffers(buffers);
        axesBuffer = buffers[0];
        cubeBuffer = buffers[1];
        teapotBuffer = buffers[2];

        // fill a buffers with geometry
        glBindBuffer(GL_ARRAY_BUFFER, axesBuffer);
        glBufferData(GL_ARRAY_BUFFER, AXES, GL_STATIC_DRAW);

        // load teapot and fill buffer with teapot data
        cube = new ObjLoader("/resources/models/cube.obj");
        try {
            cube.load();
        } catch (IOException ex) {
            Logger.getLogger(Cv4.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(1);
        }
        int length = 3 * 8 * cube.getTriangleCount();
        FloatBuffer cubeData = BufferUtils.createFloatBuffer(length);
        for (int f = 0; f < cube.getTriangleCount(); f++) {
            int[] pi = cube.getVertexIndices().get(f);
            int[] ni = cube.getNormalIndices().get(f);
            int[] ti = cube.getTexcoordIndices().get(f);
            for (int i = 0; i < 3; i++) {
                float[] position = cube.getVertices().get(pi[i]);
                float[] normal = cube.getNormals().get(ni[i]);
                float[] texcoord = cube.getTexcoords().get(ti[i]);
                cubeData.put(position);
                cubeData.put(normal);
                cubeData.put(texcoord);
            }
        }
        cubeData.rewind();
        glBindBuffer(GL_ARRAY_BUFFER, cubeBuffer);
        glBufferData(GL_ARRAY_BUFFER, cubeData, GL_STATIC_DRAW);

        // clear buffer binding, so that other code doesn't presume it (easier error detection)
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        // create a vertex array object for the geometry
        int[] arrays = new int[2];
        glGenVertexArrays(arrays);
        axesArray = arrays[0];
        cubeArray = arrays[1];

        // get axes program attributes
        int positionAttribLoc = glGetAttribLocation(axesProgram, "position");
        int colorAttribLoc = glGetAttribLocation(axesProgram, "color");
        // bind axes buffer
        glBindVertexArray(axesArray);
        glBindBuffer(GL_ARRAY_BUFFER, axesBuffer);
        // stride and offset are employed as both position and color data reside in the same vertex buffer
        glEnableVertexAttribArray(positionAttribLoc);
        glVertexAttribPointer(positionAttribLoc, 3, GL_FLOAT, false, SIZEOF_AXES_VERTEX, 0);
        glEnableVertexAttribArray(colorAttribLoc);
        glVertexAttribPointer(colorAttribLoc, 3, GL_FLOAT, false, SIZEOF_AXES_VERTEX, COLOR_OFFSET);

        // get cube program attributes
        positionAttribLoc = glGetAttribLocation(modelProgram, "position");
        int normalAttribLoc = glGetAttribLocation(modelProgram, "normal");
        int texcoordAttribLoc = glGetAttribLocation(modelProgram, "texcoord");

        // bind teapot buffer
        glBindVertexArray(cubeArray);
        glBindBuffer(GL_ARRAY_BUFFER, cubeBuffer);
        glEnableVertexAttribArray(positionAttribLoc);
        glVertexAttribPointer(positionAttribLoc, 3, GL_FLOAT, false, SIZEOF_MODEL_VERTEX, 0);
        glEnableVertexAttribArray(normalAttribLoc);
        glVertexAttribPointer(normalAttribLoc, 3, GL_FLOAT, false, SIZEOF_MODEL_VERTEX, NORMAL_OFFSET);
        glEnableVertexAttribArray(texcoordAttribLoc);
        glVertexAttribPointer(texcoordAttribLoc, 2, GL_FLOAT, false, SIZEOF_MODEL_VERTEX, TEXCOORD_OFFSET);

        // clear bindings, so that other code doesn't presume it (easier error detection)
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        glBindVertexArray(0);
    }

    private void render() {
        // Resize OpenGL viewport, i.e., the (bitmap) extents to that is the
        // OpenGL screen space [-1, 1] mapped.
        if (resized) {
            glViewport(0, 0, width, height);
            resized = false;
        }

        // animate variables
        if (animate) {
            t += 0.02f;
        }

        glPolygonMode(GL_FRONT_AND_BACK, mode);

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        Matrix4f projection = new Matrix4f()
                .perspective((float) Math.toRadians(60f), width / (float) height, 1, 500);

        Matrix4f view = new Matrix4f()
                .lookAt(camera.getEyePosition(), new Vector3f(0, 0, 0), new Vector3f(0, 1, 0));

        // Task 7:  remove drawing of the whole cube at once
        Material mat = new Material(new Vector3f(1, 1, 1), new Vector3f(1, 1, 1), new Vector3f(1, 1, 1), 45);
        drawModel(new Matrix4f(), view, projection, cubeArray, cube.getTriangleCount() * 3, mat);

        // Task 7:  draw each side of the cube with a different dice (1-6 dots) texture
        //          use second version of drawModel(...) which allows to specify an offset into VBO and the dice texture
        // Task 10: draw only back faces first and only front faces afterwards, use glCullFace(GL_FRONT [or GL_BACK])
        //              tip: draw the cube twice
        
        
    }

    private void drawModel(Matrix4f model, Matrix4f view, Matrix4f projection, int vao, int count, Material material) {
        drawModel(model, view, projection, vao, 0, count, material, 0);
    }

    private void drawModel(Matrix4f model, Matrix4f view, Matrix4f projection, int vao, int offset, int count, Material material, int diceTexture) {
        // compute model-view-projection matrix
        Matrix4f mvp = new Matrix4f(projection)
                .mul(view)
                .mul(model);

        // compute normal matrix
        Matrix3f n = model.get3x3(new Matrix3f())
                .invert()
                .transpose();

        glUseProgram(modelProgram);
        glBindVertexArray(vao); // bind vertex array to draw

        glUniform4f(lightPositionLoc, 2, 5, 3, 1);
        glUniform3f(lightAmbientColorLoc, 0.3f, 0.3f, 0.3f);
        glUniform3f(lightDiffuseColorLoc, 1, 1, 1);
        glUniform3f(lightSpecularColorLoc, 1, 1, 1);

        glUniform3f(eyePositionLoc, camera.getEyePosition().x, camera.getEyePosition().y, camera.getEyePosition().z);

        if (material != null) {
            glUniform3f(materialAmbientColorLoc, material.getAmbientColor().x, material.getAmbientColor().y, material.getAmbientColor().z);
            glUniform3f(materialDiffuseColorLoc, material.getDiffuseColor().x, material.getDiffuseColor().y, material.getDiffuseColor().z);
            glUniform3f(materialSpecularColorLoc, material.getSpecularColor().x, material.getSpecularColor().y, material.getSpecularColor().z);
            glUniform1f(materialShininessLoc, material.getShininess());
        }

        // Task 1:  set active texture to texture unit (TU) 0 using glActiveTexture(GL_TEXTURE0)
        //          bind woodTexture to TU 0 using glBindTexture(GL_TEXTURE_2D, woodTexture)
        //          assign TU 0 to woodTex sampler in GLSL program using glUniform1i(woodTexLoc, 0)
        

        // Task 7:  set active texture to texture unit (TU) 1 using glActiveTexture(GL_TEXTURE1)
        //          bind diceTexture to TU 1 using glBindTexture(GL_TEXTURE_2D, diceTexture)
        //          assign TU 1 to diceTex sampler in GLSL program using glUniform1i(diceTexLoc, 1)
        // Task 8:  set active texture to texture unit (TU) 2 using glActiveTexture(GL_TEXTURE2)
        //          bind rocksTexture to TU 2 using glBindTexture(GL_TEXTURE_2D, rocksTexture)
        //          assign TU 2 to rocksTex sampler in GLSL program using glUniform1i(rocksTexLoc, 2)
        

        FloatBuffer mvpData = BufferUtils.createFloatBuffer(16);
        FloatBuffer nData = BufferUtils.createFloatBuffer(9);
        FloatBuffer modelData = BufferUtils.createFloatBuffer(16);
        mvp.get(mvpData);
        n.get(nData);
        model.get(modelData);
        glUniformMatrix4fv(modelMvpLoc, false, mvpData); // pass MVP matrix to shader
        glUniformMatrix3fv(modelNLoc, false, nData); // pass Normal matrix to shader
        glUniformMatrix4fv(modelModelLoc, false, modelData); // pass model matrix to shader

        glDrawArrays(GL_TRIANGLES, offset, count);

        glBindTexture(GL_TEXTURE_2D, 0);
        glBindVertexArray(0);
        glUseProgram(0);
    }

    private void drawAxes(Matrix4f modelViewProjection, float length) {
        glUseProgram(axesProgram);
        glBindVertexArray(axesArray);

        FloatBuffer mvpData = BufferUtils.createFloatBuffer(16);
        modelViewProjection.get(mvpData);
        glUniform1f(axesLengthUniformLoc, length);
        glUniform1f(axesAspectUniformLoc, width / (float) height);
        glUniformMatrix4fv(axesMvpUniformLoc, false, mvpData);

        glDrawArrays(GL_LINES, 0, 6);

        glBindVertexArray(0);
        glUseProgram(0);
    }

    private int loadTexture(String filename) throws IOException {
        BufferedImage image = ImageIO.read(Cv4.class.getResourceAsStream(filename));
        byte[] pixels = ((DataBufferByte) image.getRaster().getDataBuffer()).getData();

        int internalFormat;
        int format;
        switch (image.getType()) {
            case BufferedImage.TYPE_3BYTE_BGR:
                internalFormat = GL_RGB;
                format = GL_BGR;
                break;
            case BufferedImage.TYPE_4BYTE_ABGR:
                internalFormat = GL_RGBA;
                format = GL_BGRA;
                pixels = toBGRA(pixels);
                break;
            default:
                throw new IOException("Unknown image type: " + image.getType());
        }

        ByteBuffer textureData = BufferUtils.createByteBuffer(pixels.length);
        textureData.put(pixels);
        textureData.rewind();

        int texture = 0;
        // Task 1:  create GL texture object using glGenTextures() and store it in texture local variable
        //          bind the texture using glBindTexture(GL_TEXTURE_2D, texture)
        //          upload texture data using glTexImage2D(GL_TEXTURE_2D, 0, internalFormat, <width>, <height>, 0, format, GL_UNSIGNED_BYTE, <data>)
        //              get width and height from image object
        //              data is loaded to ByteBuffer textureData
        // Task 3:  generate mipmap levels using glGenerateMipmap(GL_TEXTURE_2D)

        // Task 1:  set texture filtering using glTexParameteri(...) to GL_NEAREST
        //              minification filter: glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        //              magnification filter: glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        // Task 2:  set texture filtering using glTexParameteri(...) to GL_LINEAR
        //              minification filter: glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        //              magnification filter: glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        // Task 3:  change minification filter (GL_TEXTURE_MIN_FILTER) from GL_LINEAR to GL_LINEAR_MIPMAP_LINEAR
        // Task 5:  set texture wrap mode to GL_MIRRORED_REPEAT in both S and T directions
        //              S direction: glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT)
        //              T direction: glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT)
        //              also try other modes, listed in the attached PDF :)
        

        // unbind texture
        glBindTexture(GL_TEXTURE_2D, 0);

        return texture;
    }

    private byte[] toBGRA(byte[] abgr) {
        byte[] bgra = new byte[abgr.length];
        for (int i = 0; i < abgr.length; i += 4) {
            bgra[i] = abgr[i + 1];
            bgra[i + 1] = abgr[i + 2];
            bgra[i + 2] = abgr[i + 3];
            bgra[i + 3] = abgr[i];
        }
        return bgra;
    }

    private int loadShader(String filename, int shaderType) throws IOException {
        String source = readAllFromResource(filename);
        int shader = glCreateShader(shaderType);

        // create and compile GLSL shader
        glShaderSource(shader, source);
        glCompileShader(shader);

        // check GLSL shader compile status
        int status = glGetShaderi(shader, GL_COMPILE_STATUS);
        if (status == GL_FALSE) {
            String log = glGetShaderInfoLog(shader);
            System.err.println(log);
        }

        return shader;
    }

    private int loadProgram(String vertexShaderFile, String fragmentShaderFile) throws IOException {
        // load vertex and fragment shaders (GLSL)
        int vs = loadShader(vertexShaderFile, GL_VERTEX_SHADER);
        int fs = loadShader(fragmentShaderFile, GL_FRAGMENT_SHADER);

        // create GLSL program, attach shaders and compile it
        int program = glCreateProgram();
        glAttachShader(program, vs);
        glAttachShader(program, fs);
        glLinkProgram(program);

        int status = glGetProgrami(program, GL_LINK_STATUS);
        if (status == GL_FALSE) {
            String log = glGetProgramInfoLog(program);
            System.err.println(log);
        }

        return program;
    }

    private String readAllFromResource(String resource) throws IOException {
        InputStream is = Cv4.class.getResourceAsStream(resource);
        if (is == null) {
            throw new IOException("Resource not found: " + resource);
        }

        BufferedReader reader = new BufferedReader(new InputStreamReader(is));
        StringBuilder sb = new StringBuilder();

        int c;
        while ((c = reader.read()) != -1) {
            sb.append((char) c);
        }

        return sb.toString();
    }

}
